# Copyright 2022 Q-CTRL. All rights reserved.
#
# Licensed under the Q-CTRL Terms of service (the "License"). Unauthorized
# copying or use of this file, via any medium, is strictly prohibited.
# Proprietary and confidential. You may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#    https://q-ctrl.com/terms
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS. See the
# License for the specific language.
# pylint:disable=missing-module-docstring
from typing import (
    Any,
    Dict,
)

from qctrl.builders.graphql_utils.handlers import NamedObjectHandler
from qctrl.builders.type_registry import TypeOverrideMixin


class ComplexNumberHandler(
    NamedObjectHandler, TypeOverrideMixin
):  # pylint:disable=too-many-ancestors
    """Custom handler for ComplexNumber."""

    _type_hint = complex
    _object_name = "ComplexNumber"

    def process_loaded_data(self, data: Dict[str, Any]):
        return complex(data.get("real") or 0, data.get("imag") or 0)
